from typing import Any
from .style import style

DEFAULT_SYNTAX_HIGHLIGHT_COLORS = {
    'key': '#e06c75',
    'arr': '#ffd700',
    'dict': '#ffd700',
    'str': '#98c379',
    'num': '#d19a58',
    'bool': '#d19a66',
}

NEWLINE = '\n'
TAB = '\t'


def highlight(value: Any,
              indent: int = 2,
              curr_indent: int = 0,
              *,
              trailing_comma: bool = False,
              pre_indent: bool = True,
              colors: dict[str, str] | None = None):
    """Converts a value to a string with syntax highlighting.

    Args:
        val (any): Value to highlight. Can be a dict, list, str, int, bool, etc.
        indent (int, optional): Amount of spaces to indent.
        curr_indent (int, optional): Current indent level.
        trailing_comma (bool, optional): Whether to add a trailing comma to the last item in a list or dict.
        pre_indent (bool, optional): Whether to indent the first line of a list or dict.
        colors (dict[str, str], optional): Colors to use for syntax highlighting.
            Uses `DEFAULT_SYNTAX_HIGHLIGHT_COLORS` if None.

    Returns:
        str: Values as a string with syntax highlighting.
    """
    if colors is None:
        colors = DEFAULT_SYNTAX_HIGHLIGHT_COLORS
    if isinstance(value, dict):
        return pretty_dict(value,
                           indent=indent,
                           curr_indent=curr_indent,
                           trailing_comma=trailing_comma,
                           pre_indent=pre_indent,
                           colors=colors)
    if isinstance(value, list):
        return pretty_list(value,
                           indent=indent,
                           curr_indent=curr_indent,
                           trailing_comma=trailing_comma,
                           pre_indent=pre_indent,
                           colors=colors)
    if isinstance(value, str):
        return style(repr(value), fg=colors['str'])
    if isinstance(value, int) or isinstance(value, float):
        return style(repr(value), fg=colors['num'])
    if isinstance(value, bool):
        return style(repr(value), fg=colors['bool'])
    return repr(value)


def pretty_dict(dictionary: dict,
                indent: int = 2,
                curr_indent: int = 0,
                *,
                trailing_comma: bool = False,
                pre_indent: bool = True,
                colors: dict[str, str] | None = None):
    """Converts a dict to a string with syntax highlighting.

    Args:
        dictionary (dict): Dictionary to syntax highlight.
        indent (int, optional): Amount of spaces to indent.
        curr_indent (int, optional): Current indent level.
        trailing_comma (bool, optional): Whether to add a trailing comma to the last item in a list or dict.
        pre_indent (bool, optional): Whether to indent the first line of a list or dict.
        colors (dict[str, str], optional): Colors to use for syntax highlighting.
            Uses `DEFAULT_SYNTAX_HIGHLIGHT_COLORS` if None.

    Returns:
       str: Dictionary as a string with syntax highlighting.
    """
    if colors is None:
        colors = DEFAULT_SYNTAX_HIGHLIGHT_COLORS

    lcub = style('{', fg=colors['dict'])
    rcub = style('}', fg=colors['dict'])

    pre_identation = " " * curr_indent if pre_indent else ''
    output_str = f'{pre_identation}{lcub}{NEWLINE}'

    for i, (key, val) in enumerate(dictionary.items()):
        pretty_key = style(repr(key), fg=colors['key'])
        pretty_value = highlight(val,
                                 indent=indent,
                                 curr_indent=indent + curr_indent,
                                 pre_indent=False,
                                 colors=colors)

        comma = ',' if (i < len(dictionary) - 1) else ''
        output_str += f'{" "*(curr_indent + indent)}{pretty_key}: {pretty_value}{comma}{NEWLINE}'

    output_str += f'{" "* curr_indent}{rcub}{ "," + NEWLINE if trailing_comma else ""}'
    return output_str


def pretty_list(lst: list,
                indent: int = 2,
                curr_indent: int = 0,
                *,
                trailing_comma: bool = False,
                pre_indent: bool = True,
                colors: dict[str, str] | None = None):
    """Converts a list to a string with syntax highlighting.

    Args:
        lst (list): List to syntax highlight.
        indent (int, optional): Amount of spaces to indent.
        curr_indent (int, optional): Current indent level.
        trailing_comma (bool, optional): Whether to add a trailing comma to the last item in a list or dict.
        pre_indent (bool, optional): Whether to indent the first line of a list or dict.
        colors (dict[str, str] | None, optional): Colors to use for syntax highlighting.
            Uses `DEFAULT_SYNTAX_HIGHLIGHT_COLORS` if None.

    Returns:
        str: List as a string with syntax highlighting.
    """
    if colors is None:
        colors = DEFAULT_SYNTAX_HIGHLIGHT_COLORS

    lsqb = style('[', fg=colors['arr'])
    rsqb = style(']', fg=colors['arr'])

    pre_identation = " " * curr_indent if pre_indent else ''
    output_str = f'{pre_identation}{lsqb}{NEWLINE}'

    for i, val in enumerate(lst):
        pretty_value = highlight(val,
                                 indent=indent,
                                 curr_indent=indent + curr_indent,
                                 pre_indent=False,
                                 colors=colors)
        comma = ',' if (i < len(lst) - 1) else ''
        output_str += f'{" "*(curr_indent + indent)}{pretty_value}{comma}{NEWLINE}'

    output_str += f'{" "* curr_indent}{rsqb}{"," + NEWLINE if trailing_comma else ""}'
    return output_str



from enum import Enum, IntEnum

from .syntax import TAB

ESC = '\033['
RESET = f'{ESC}0m'


class Fg(Enum):
    """ANSI escape codes for foreground colors.
    """
    DEFAULT = 39
    BLACK = 30
    RED = 31
    GREEN = 32
    YELLOW = 33
    BLUE = 34
    MAGENTA = 35
    CYAN = 36
    WHITE = 37
    GRAY = 90
    BRIGHT_RED = 91
    BRIGHT_GREEN = 92
    BRIGHT_YELLOW = 93
    BRIGHT_BLUE = 94
    BRIGHT_MANGENTA = 95
    BRIGHT_CYAN = 96
    BRIGHT_WHITE = 97


class Bg(Enum):
    """ANSI escape codes for background colors.
    """
    DEFAULT = 49
    BLACK = 40
    RED = 41
    GREEN = 42
    YELLOW = 43
    BLUE = 44
    MAGENTA = 45
    CYAN = 46
    WHITE = 47
    GRAY = 100
    BRIGHT_RED = 101
    BRIGHT_GREEN = 102
    BRIGHT_YELLOW = 103
    BRIGHT_BLUE = 104
    BRIGHT_MANGENTA = 105
    BRIGHT_CYAN = 106
    BRIGHT_WHITE = 107


class Frmt(IntEnum):
    """ANSI escape codes for text formatting.
    """
    DEFAULT = 0
    BOLD = 1
    FAINT = 2
    ITALIC = 3
    UNDERLINE = 4
    BLINK_SLOW = 5
    BLINK_FAST = 6
    INVERT = 7
    HIDE = 8
    STRIKETHROUGH = 9
    DOUBLE_UNDERLINE = 21
    OVERLINE = 53


def fg_rgb(rgb_color: tuple[int, int, int]) -> str:
    """Converts an RGB tuple to an ANSI escape code for foreground color.

    Args:
        rgb_color (tuple[int, int, int]): Tuple of RGB values as integers.

    Returns:
        str: ANSI escape code for foreground color.
    """
    return f'38;2;{rgb_color[0]};{rgb_color[1]};{rgb_color[2]}'


def fg_hex(hex_color: str) -> str:
    """Converts a hex color string to an ANSI escape code for foreground color.

    Args:
        hex_color (str): Hex color string.

    Returns:
        str: ANSI escape code for foreground color.
    """
    return fg_rgb(hex2rgb(hex_color))


def bg_rgb(rgb_color: tuple[int, int, int]) -> str:
    """Converts an RGB tuple to an ANSI escape code for background color.

    Args:
        rgb_color (tuple[int, int, int]): Tuple of RGB values as integers.

    Returns:
        str: ANSI escape code for background color.
    """
    return f'48;2;{rgb_color[0]};{rgb_color[1]};{rgb_color[2]}'


def bg_hex(hex_color: str):
    """Converts a hex color string to an ANSI escape code for background color.

    Args:
        hex_color (str): Hex color string.

    Returns:
        str: ANSI escape code for background color.
    """
    return bg_rgb(hex2rgb(hex_color))


def hex2rgb(hex_color: str) -> tuple[int, int, int]:
    """Converts a hex color string to an RGB tuple of integers.

    Args:
        hex_color (str): Hex color string.

    Returns:
        tuple[int, int, int]: RGB tuple of integers.
    """

    def hex2dec(hex_str: str) -> int:
        """Converts a hex string to an base 10 integer.

        Args:
            hex_str (str): Hex string.

        Returns:
            int: Base 10 integer.
        """
        return int(hex_str, 16)

    if hex_color[0] == '#':
        hex_color = hex_color[1:]

    r = hex2dec(hex_color[0:2])
    g = hex2dec(hex_color[2:4])
    b = hex2dec(hex_color[4:6])

    return (r, g, b)


def print_styles(ansi_type: type[Fg] | type[Bg] | type[Frmt]) -> None:
    """Prints all common ANSI escape codes for a given type.
    """

    for style in ansi_type:
        print(f'{ESC}{style.value}mExample Text{TAB}{RESET}({style.name}) ')


def print_styles_all() -> None:
    """Prints all common ANSI escape codes.
    """
    print_styles(Fg)
    print()
    print_styles(Bg)
    print()
    print_styles(Frmt)



from typing import Union
from .ansi import Fg, Bg, Frmt, ESC, RESET, bg_hex, bg_rgb, fg_hex, fg_rgb


def style(text: str,
          fg: Fg | str | tuple[int, int, int] |
          list[Union[int, int, int]] = Fg.DEFAULT,
          bg: Bg | str | tuple[int, int, int] |
          list[Union[int, int, int]] = Bg.DEFAULT,
          frmt: list[Frmt] | None = None) -> str:
    """ Styles string using ANSI escape codes.

    Args:
        text (str): String to style.
        fg (Fg | str | tuple[int, int, int] | list[int, int, int], optional): Foreground color.
            Either a Fg enum, a hex color string, or an RGB tuple or list.
        bg (Bg | str | tuple[int, int, int] | list[int, int, int], optional): Background color.
            Either a Bg enum, a hex color string, or an RGB tuple or list.
        frmt (Frmt[Style], optional): List of formatting to apply.

    Returns:
        str: Styled text.
    """

    if isinstance(fg, str):
        fore = fg_hex(fg)
    if isinstance(fg, tuple):
        fore = fg_rgb(fg)
    if isinstance(fg, Fg):
        fore = f'{fg.value}'

    if isinstance(bg, str):
        back = bg_hex(bg)
    if isinstance(bg, tuple):
        back = bg_rgb(bg)
    if isinstance(bg, Bg):
        back = f'{bg.value}'

    if frmt is None:
        frmt = []
    else:
        frmt.sort()

    frmt_string = ''
    for f in frmt:
        frmt_string += f';{f.value}'

    return f'{ESC}{fore};{back}{frmt_string}m{text}{RESET}'



from itertools import product, permutations as _permutations
from typing import Generator


def permutations(string: str,
                 max_len: int | None = None,
                 min_len: int = 1,
                 *,
                 allow_repetition: bool = False) -> Generator[str, None, None]:
    """Generates all permutations of a string.

    Args:
        string (str): String to permute.
        max_len (int, optional): Maximum length of permutations. Defaults to length of the string.
        min_len (int, optional): Minimum length of permutations. Defaults to 1.
        allow_repetition (bool, optional): Whether to allow repetition of characters. Defaults to False.

    Yields:
        str: Permutations of the string.
    """
    if max_len is None:
        max_len = len(string)
    else:
        max_len = min(max_len, len(string))

    min_len = min(min_len, max_len)

    if min_len <= 0:
        min_len = 1

    if allow_repetition:
        for i in range(min_len, max_len + 1):
            for p in product(string, repeat=max_len):
                yield ''.join(p)
        return

    for i in range(min_len, max_len + 1):
        for p in _permutations(string, i):
            yield ''.join(p)
